/* checksum.S: Sparc optimized checksum code.
 *
 *  Copyright(C) 1995 Linus Torvalds
 *  Copyright(C) 1995 Miguel de Icaza
 *  Copyright(C) 1996 David S. Miller
 *
 * derived from:
 *	Linux/Alpha checksum c-code
 *      Linux/ix86 inline checksum assembly
 *      RFC1071 Computing the Internet Checksum (esp. Jacobsons m68k code)
 *	David Mosberger-Tang for optimized reference c-code
 *	BSD4.4 portable checksum routine
 */

#include <asm-sparc/cprefix.h>

	.text
	.align 4

	/* Every effort has been made to avoid allocating a new
	 * window as that would defeat the entire purpose of
	 * writing this junk in assembler.
	 *
	 * For the purposes of optimization heuristics we assume
	 * an average cache line size of 32 bytes.
	 */

	/* Rules for csum_partial*():
	 * 1) Arrange all branches assuming the cpu predicts
	 *    them to be taken.
	 * 2) Absolutely _NO_ loads/stores in delay slots,
	 *    with no exceptions!  If you don't understand
	 *    why you need to abide by this then you shouldn't
	 *    be touching this code, QED.
	 */

	/* The scheme of things is, we assume a cache miss on the
	 * first load and the second load can not pair with the
	 * first one anyways.  Regardless of what happens to
	 * the first load the second one will indeed pair every
	 * time with the add, and the same goes for each ld/addxcc
	 * duo as well.  We eat one dead cycle at the end to catch
	 * up with the missed addxcc at the top.  Note how we work
	 * the cache backwards, this is the only way to guarentee
	 * a real hit in all subsequent loads for all Sparc cache
	 * types.
	 */
#define CSUM_BIGCHUNK_FIRST(ptr, accum, offset, tmp1, tmp2) \
	ld	[%ptr + offset + 0x1c], %tmp1;        \
	ld	[%ptr + offset + 0x18], %tmp2;        \
	addcc	%tmp1, %accum, %accum;                \
	ld	[%ptr + offset + 0x14], %tmp1;        \
	addxcc	%tmp2, %accum, %accum;                \
	ld	[%ptr + offset + 0x10], %tmp2;        \
	addxcc	%tmp1, %accum, %accum;                \
	ld	[%ptr + offset + 0x0c], %tmp1;        \
	addxcc	%tmp2, %accum, %accum;                \
	ld	[%ptr + offset + 0x08], %tmp2;        \
	addxcc	%tmp1, %accum, %accum;                \
	ld	[%ptr + offset + 0x04], %tmp1;        \
	addxcc	%tmp2, %accum, %accum;                \
	ld	[%ptr + offset + 0x00], %tmp2;        \
	addxcc	%tmp1, %accum, %accum;                \
	addxcc	%tmp2, %accum, %accum;
	
#define CSUM_BIGCHUNK(ptr, accum, offset, tmp1, tmp2) \
	ld	[%ptr + offset + 0x1c], %tmp1;        \
	ld	[%ptr + offset + 0x18], %tmp2;        \
	addxcc	%tmp1, %accum, %accum;                \
	ld	[%ptr + offset + 0x14], %tmp1;        \
	addxcc	%tmp2, %accum, %accum;                \
	ld	[%ptr + offset + 0x10], %tmp2;        \
	addxcc	%tmp1, %accum, %accum;                \
	ld	[%ptr + offset + 0x0c], %tmp1;        \
	addxcc	%tmp2, %accum, %accum;                \
	ld	[%ptr + offset + 0x08], %tmp2;        \
	addxcc	%tmp1, %accum, %accum;                \
	ld	[%ptr + offset + 0x04], %tmp1;        \
	addxcc	%tmp2, %accum, %accum;                \
	ld	[%ptr + offset + 0x00], %tmp2;        \
	addxcc	%tmp1, %accum, %accum;                \
	addxcc	%tmp2, %accum, %accum;
	
	/* We keep track of _minimal_ state.  Note the trick
	 * where we only modify the len parameter to become aligned
	 * at the beginning and that is it!  For the rest, simple
	 * srl into %g7 for the big unrolled 128byte loop and
	 * low order bit tests on the len value for the remaining
	 * expansions work just fine and minimize starvation on the
	 * register ports of the chip.
	 *
	 * Experiments were done with unrolling the loops to >128 byte
	 * chunks with no noticable performance gain.
	 */

	.align 4
	.globl	C_LABEL(csum_partial)
C_LABEL(csum_partial):		/* %o0=buff %o1=len %o2=sum */
	/* Align on 4-byte boundry if we need to. */
	andcc	%o0, 2, %g0
	be	2f
	 cmp	%o1, 32

	subcc	%o1, 2, %o1
	lduh	[%o0], %o4
	bge,a	1f
	 add	%o0, 2, %o0

	b	4f
	add	%o1, 2, %o1
1:
	sll	%o4, 16, %o4
	addcc	%o4, %o2, %o2
	srl	%o2, 16, %o5
	addx	%g0, %o5, %o5
	sll	%o2, 16, %o2
	sll	%o5, 16, %o5
	srl	%o2, 16, %o2
	or	%o5, %o2, %o2
	/* Now that we are longword aligned we go for 32-byte
	 * alignment to make the cache trickles in the unrolled
	 * loops stall only twice per macro expansion.  But first see
	 * if at least 32 bytes are left in buff.  These instructions
	 * to get ourselves aligned are _petty_ cycles compared to the
	 * kind of cycles we eat below in the unrolled loops if
	 * we are not cache line aligned.
	 *
	 * XXX expand this alignment loop into one sequence which looks
	 * XXX like
	 * XXX
	 * XXX      set forward_label, %reg
	 * XXX      sll longs_to_go, 3, %offset
	 * XXX      jmpl %reg + %offset, %g0
	 * XXX       addcc %g0, %g0, %g0	! clear %psr.Carry
	 * XXX forward_label:
	 * XXX      bunch 'o' loads/n/stores/addxcc's
	 * XXX      addx %g0, %g0, %g0
	 * XXX big_128byte_loop:
	 */
	cmp	%o1, 32
2:
	bg	7f
	 andcc	%o0, 0x1c, %g7

	b	8f
	 andcc	%o1, 0x20, %g0

7:
	bne	1f
	 srl	%g7, 2, %g7

	b	9f
	 srl	%o1, 7, %g7

1:
	ld	[%o0 + 0x00], %o4
	addcc	%o4, %o2, %o2
	sub	%o1, 0x4, %o1
	addx	%g0, %o2, %o2
	subcc	%g7, 1, %g7
	bne	1b
	 add	%o0, 4, %o0

0:
	srl	%o1, 7, %g7
9:
	orcc	%g0, %g7, %g0
	be	2f
	 andcc	%o1, 0x40, %g0

	/* CSUM 128 bytes at a time */
1:
	CSUM_BIGCHUNK_FIRST(o0, o2, 0x00, o4, o5)
	CSUM_BIGCHUNK(o0, o2, 0x20, o4, o5)
	CSUM_BIGCHUNK(o0, o2, 0x40, o4, o5)
	CSUM_BIGCHUNK(o0, o2, 0x60, o4, o5)
	addx	%g0, %o2, %o2
	subcc	%g7, 1, %g7
	bnz	1b
	 add	%o0, 128, %o0

	andcc	%o1, 0x40, %g0
2:
	be	8f
	 andcc	%o1, 0x20, %g0

	/* CSUM 64 byte end cruft. */
1:
	CSUM_BIGCHUNK_FIRST(o0, o2, 0x00, o4, o5)
	CSUM_BIGCHUNK(o0, o2, 0x20, o4, o5)
	add	%o0, 64, %o0
	addx	%g0, %o2, %o2

	andcc	%o1, 0x20, %g0
8:
	be	2f
	 andcc	%o1, 0x1c, %g7

	/* CSUM 32 byte end cruft.  It makes no sense to unroll smaller
	 * powers of two.
	 */
1:
	CSUM_BIGCHUNK_FIRST(o0, o2, 0x00, o4, o5)
	add	%o0, 32, %o0
	addx	%g0, %o2, %o2

	andcc	%o1, 0x1c, %g7
2:
	be	4f
	 srl	%g7, 2, %g7

3:
	ld	[%o0 + 0x00], %o4
	addcc	%o4, %o2, %o2
	addx	%g0, %o2, %o2
	subcc	%g7, 1, %g7
	bne	3b
	 add	%o0, 4, %o0

4:
	andcc	%o1, 3, %o1
	be	7f
	 cmp	%o1, 1

	bne,a	2f
	 subcc	%o1, 2, %o1

	b	5f
	 clr	%o4

2:
	lduh	[%o0 + 0x00], %o4
	add	%o0, 2, %o0
	be,a	6f
	 addcc	%o4, %o2, %o2

	sll	%o4, 16, %o4
5:
	ldub	[%o0 + 0x00], %o5
	sll	%o5, 8, %o5
	or	%o5, %o4, %o4
	addcc	%o4, %o2, %o2
6:
	addx	%g0, %o2, %o2
7:
	retl
	 mov	%o2, %o0

	/* The time between the third ldd and the following ldd must be
	 * as short as possible.  We want to take no risk that any of the
	 * stores in between (or an IRQ) blow the cache line we worked so
	 * hard to bring in.
	 *
	 * Experimentation with doing a ld[%dest], %g0 at the top before
	 * the first store showed it to be a dead cycle and actually hurt
	 * performance.
	 */
#define CSUMCOPY_BIGCHUNK_FIRST(src, dest, accum, offset, t0, t1, t2, t3, t4, t5) \
	ldd	[%src + offset + 0x18], %t0; \
	ldd	[%src + offset + 0x10], %t2; \
	ldd	[%src + offset + 0x08], %t4; \
	st	%t0, [%dest + offset + 0x18]; \
	addcc	%t0, %accum, %accum; \
	st	%t1, [%dest + offset + 0x1c]; \
	addxcc	%t1, %accum, %accum; \
	st	%t2, [%dest + offset + 0x10]; \
	addxcc	%t2, %accum, %accum; \
	st	%t3, [%dest + offset + 0x14]; \
	addxcc	%t3, %accum, %accum; \
	ldd	[%src + offset + 0x00], %t0; \
	st	%t4, [%dest + offset + 0x08]; \
	addxcc	%t4, %accum, %accum; \
	st	%t5, [%dest + offset + 0x0c]; \
	addxcc	%t5, %accum, %accum; \
	st	%t0, [%dest + offset + 0x00]; \
	addxcc	%t0, %accum, %accum; \
	st	%t1, [%dest + offset + 0x04]; \
	addxcc	%t1, %accum, %accum; \

#define CSUMCOPY_BIGCHUNK(src, dest, accum, offset, t0, t1, t2, t3, t4, t5) \
	ldd	[%src + offset + 0x18], %t0; \
	ldd	[%src + offset + 0x10], %t2; \
	ldd	[%src + offset + 0x08], %t4; \
	st	%t0, [%dest + offset + 0x18]; \
	addxcc	%t0, %accum, %accum; \
	st	%t1, [%dest + offset + 0x1c]; \
	addxcc	%t1, %accum, %accum; \
	st	%t2, [%dest + offset + 0x10]; \
	addxcc	%t2, %accum, %accum; \
	st	%t3, [%dest + offset + 0x14]; \
	addxcc	%t3, %accum, %accum; \
	ldd	[%src + offset + 0x00], %t0; \
	st	%t4, [%dest + offset + 0x08]; \
	addxcc	%t4, %accum, %accum; \
	st	%t5, [%dest + offset + 0x0c]; \
	addxcc	%t5, %accum, %accum; \
	st	%t0, [%dest + offset + 0x00]; \
	addxcc	%t0, %accum, %accum; \
	st	%t1, [%dest + offset + 0x04]; \
	addxcc	%t1, %accum, %accum; \

	.align 4
	.globl	C_LABEL(csum_partial_copy)
C_LABEL(csum_partial_copy):	/* %o0=src %o1=dst %o2=len %o3=sum */
	/* Align source on 4-byte boundry if we need to. */
	andcc	%o0, 2, %g0
	be	2f
	 srl	%o2, 7, %g7

	subcc	%o2, 2, %o2
	lduh	[%o0], %o4
	srl	%o2, 7, %g7
	bge,a	1f
	 add	%o0, 2, %o0

	b	4f
	add	%o2, 2, %o2
1:
	sth	%o4, [%o1]
	add	%o1, 2, %o1
	sll	%o4, 16, %o4
	addcc	%o4, %o3, %o3
	srl	%o3, 16, %o5
	addx	%g0, %o5, %o5
	sll	%o3, 16, %o3
	sll	%o5, 16, %o5
	srl	%o3, 16, %o3
	or	%o5, %o3, %o3
2:
	/* Align source on 32 byte boundry if we need to and can. */
	cmp	%o2, 32
	ble	8f
	 andcc	%o2, 0x20, %g0

	andcc	%o0, 0x1c, %g7
	be	0f
	 srl	%g7, 2, %g7

1:
	ld	[%o0], %o4
	add	%o0, 4, %o0
	addcc	%o4, %o3, %o3
	st	%o4, [%o1]
	addx	%g0, %o3, %o3
	sub	%o2, 4, %o2
	subcc	%g7, 1, %g7
	bne	1b
	 add	%o1, 4, %o1

0:
	srl	%o2, 7, %g7
9:
	orcc	%g0, %g7, %g0
2:
	be	2f
	 andcc	%o2, 0x40, %g0

	/* CSUM 128 bytes at a time */
1:
	CSUMCOPY_BIGCHUNK_FIRST(o0, o1, o3, 0x00, o4, o5, g2, g3, g4, g5)
	CSUMCOPY_BIGCHUNK(o0, o1, o3, 0x20, o4, o5, g2, g3, g4, g5)
	CSUMCOPY_BIGCHUNK(o0, o1, o3, 0x40, o4, o5, g2, g3, g4, g5)
	CSUMCOPY_BIGCHUNK(o0, o1, o3, 0x60, o4, o5, g2, g3, g4, g5)
	addx	%g0, %o3, %o3
	add	%o0, 128, %o0
	subcc	%g7, 1, %g7
	bnz	1b
	 add	%o1, 128, %o1

	andcc	%o2, 0x40, %g0
2:
	be	8f
	 andcc	%o2, 0x20, %g0

	/* CSUM 64 byte end cruft. */
1:
	CSUMCOPY_BIGCHUNK_FIRST(o0, o1, o3, 0x00, o4, o5, g2, g3, g4, g5)
	CSUMCOPY_BIGCHUNK(o0, o1, o3, 0x20, o4, o5, g2, g3, g4, g5)
	add	%o0, 64, %o0
	addx	%g0, %o3, %o3
	add	%o1, 64, %o1

	andcc	%o2, 0x20, %g0
8:
	be	2f
	 andcc	%o2, 0x1c, %g7

	/* CSUM 32 byte end cruft.  It makes no sense to unroll smaller
	 * powers of two.
	 */
1:
	CSUMCOPY_BIGCHUNK_FIRST(o0, o1, o3, 0x00, o4, o5, g2, g3, g4, g5)
	add	%o0, 32, %o0
	addx	%g0, %o3, %o3
	add	%o1, 32, %o1

	andcc	%o2, 0x1c, %g7
2:
	be	4f
	 srl	%g7, 2, %g7

3:
	ld	[%o0 + 0x00], %o4
	addcc	%o4, %o3, %o3
	st	%o4, [%o1 + 0x00]
	addx	%g0, %o3, %o3
	subcc	%g7, 1, %g7
	add	%o1, 4, %o1
	bne	3b
	 add	%o0, 4, %o0

4:
	andcc	%o2, 3, %o2
	be	7f
	 cmp	%o2, 1

	bne,a	2f
	 subcc	%o2, 2, %o2

	b	5f
	 clr	%o4

2:
	lduh	[%o0 + 0x00], %o4
	sth	%o4, [%o1 + 0x00]
	add	%o0, 2, %o0
	add	%o1, 2, %o1
	be,a	6f
	 addcc	%o4, %o3, %o3

	sll	%o4, 16, %o4
5:
	ldub	[%o0 + 0x00], %o5
	stb	%o5, [%o1 + 0x00]	! XXX this is a _really_ bad stall
	sll	%o5, 8, %o5
	or	%o5, %o4, %o4
	addcc	%o4, %o3, %o3
6:
	addx	%g0, %o3, %o3
7:
	retl
	 mov	%o3, %o0

	/* yeah this shit coming up is ugly, just be quiet and
	 * be glad you didn't have to write and test it...
	 */

	/* All the fixup routines expect %g6=trap_address and
	 * %g1=imm_offset if any.  You _CANNOT_ modify the
	 * condition codes or else this _WILL NOT_ work!!!
	 */

#define CSUM_FIXUP_LD(dreg, basereg, offreg, retpc, t0, t1, t2) \
	add	%offreg, %basereg, %offreg; \
	ldub	[%offreg + 0x3], %dreg;     \
	ldub	[%offreg + 0x2], %t0;       \
	ldub	[%offreg + 0x1], %t1;       \
	sll	%t0, 8, %t0;                \
	ldub	[%offreg + 0x0], %t2;       \
	sll	%t1, 16, %t1;               \
	or	%t0, %dreg, %dreg;          \
	sll	%t2, 24, %t2;               \
	or	%t1, %dreg, %dreg;          \
	jmp	%retpc + 0x4;               \
	 or	%t2, %dreg, %dreg;          \

	/* We ignore the offset here. */
#define CSUM_FIXUP_LDUH(dreg, breg, offreg, retpc) \
	ldub	[%breg + 0x1], %dreg;     \
	ldub	[%breg + 0x0], %offreg;   \
	sll	%offreg, 8, %offreg;        \
	jmp	%retpc + 0x4;               \
	 or	%offreg, %dreg, %dreg;      \

	.align 4
	.globl	C_LABEL(csum_partial_ldo4fixup)
C_LABEL(csum_partial_ldo4fixup):
	CSUM_FIXUP_LD(o4, o0, g1, g6, g2, g3, g4)

	.align 4
	.globl	C_LABEL(csum_partial_ldo5fixup)
C_LABEL(csum_partial_ldo5fixup):
	CSUM_FIXUP_LD(o5, o0, g1, g6, g2, g3, g4)

	.align 4
	.globl	C_LABEL(csum_partial_lduho4fixup)
C_LABEL(csum_partial_lduho4fixup):
	CSUM_FIXUP_LDUH(o4, o0, g1, g6)

	/* The csum_partial_copy fixups are more numerous
	 * as we have fewer registers to work with.
	 * yeeee....
	 */

	/* This macro _MUST_ expand to 32-instructions==128-bytes. */
#define CCOPY_FIXUP_LDD(dreg0, dreg1, breg, offset, retpc, tmp) \
	ldub	[%breg + offset + 0x3], %dreg0; \
	ldub	[%breg + offset + 0x2], %tmp;   \
	sll	%tmp, 8, %tmp;                  \
	or	%tmp, %dreg0, %dreg0;           \
	ldub	[%breg + offset + 0x1], %tmp;   \
	sll	%tmp, 16, %tmp;                 \
	or	%tmp, %dreg0, %dreg0;           \
	ldub	[%breg + offset + 0x0], %tmp;   \
	sll	%tmp, 24, %tmp;                 \
	or	%tmp, %dreg0, %dreg0;           \
	ldub	[%breg + offset + 0x7], %dreg1; \
	ldub	[%breg + offset + 0x6], %tmp;   \
	sll	%tmp, 8, %tmp;                  \
	or	%tmp, %dreg1, %dreg1;           \
	ldub	[%breg + offset + 0x5], %tmp;   \
	sll	%tmp, 16, %tmp;                 \
	or	%tmp, %dreg1, %dreg1;           \
	ldub	[%breg + offset + 0x4], %tmp;   \
	sll	%tmp, 24, %tmp;                 \
	jmp	%retpc + 0x4;                   \
	 or	%tmp, %dreg1, %dreg1;           \
	nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; \

	.align 4
	.globl	C_LABEL(csum_copy_lddo4o5fixups)
	.globl	C_LABEL(csum_copy_lddg2g3fixups)
	.globl	C_LABEL(csum_copy_lddg4g5fixups)
C_LABEL(csum_copy_lddo4o5fixups):
	CCOPY_FIXUP_LDD(o4, o5, o0, 0x00, g6, g1) /* 0x000 */
	CCOPY_FIXUP_LDD(o4, o5, o0, 0x18, g6, g1) /* 0x080 */
	CCOPY_FIXUP_LDD(o4, o5, o0, 0x20, g6, g1) /* 0x100 */
	CCOPY_FIXUP_LDD(o4, o5, o0, 0x38, g6, g1) /* 0x180 */
	CCOPY_FIXUP_LDD(o4, o5, o0, 0x40, g6, g1) /* 0x200 */
	CCOPY_FIXUP_LDD(o4, o5, o0, 0x58, g6, g1) /* 0x280 */
	CCOPY_FIXUP_LDD(o4, o5, o0, 0x60, g6, g1) /* 0x300 */
	CCOPY_FIXUP_LDD(o4, o5, o0, 0x78, g6, g1) /* 0x380 */
C_LABEL(csum_copy_lddg2g3fixups):
	CCOPY_FIXUP_LDD(g2, g3, o0, 0x10, g6, g1) /* 0x000 */
	CCOPY_FIXUP_LDD(g2, g3, o0, 0x30, g6, g1) /* 0x080 */
	CCOPY_FIXUP_LDD(g2, g3, o0, 0x50, g6, g1) /* 0x100 */
	CCOPY_FIXUP_LDD(g2, g3, o0, 0x70, g6, g1) /* 0x180 */
C_LABEL(csum_copy_lddg4g5fixups):
	CCOPY_FIXUP_LDD(g4, g5, o0, 0x08, g6, g1) /* 0x000 */
	CCOPY_FIXUP_LDD(g4, g5, o0, 0x28, g6, g1) /* 0x080 */
	CCOPY_FIXUP_LDD(g4, g5, o0, 0x48, g6, g1) /* 0x100 */
	CCOPY_FIXUP_LDD(g4, g5, o0, 0x68, g6, g1) /* 0x180 */

	/* Immediate offset is always zero. */
#define CCOPY_FIXUP_LD(dreg, breg, retpc, tmp) \
	ldub	[%breg + 0x3], %dreg;     \
	ldub	[%breg + 0x2], %tmp;      \
	sll	%tmp, 8, %tmp;            \
	or	%tmp, %dreg, %dreg;       \
	ldub	[%breg + 0x1], %tmp;      \
	sll	%tmp, 16, %tmp;           \
	or	%tmp, %dreg, %dreg;       \
	ldub	[%breg + 0x0], %tmp;      \
	sll	%tmp, 24, %tmp;           \
	jmp	%retpc + 0x4;             \
	 or	%tmp, %dreg, %dreg;       \

#define CCOPY_FIXUP_LDUH(dreg, breg, retpc, tmp) \
	ldub	[%breg + 0x1], %dreg;     \
	ldub	[%breg + 0x0], %tmp;      \
	sll	%tmp, 8, %tmp;            \
	jmp	%retpc + 0x4;             \
	 or	%tmp, %dreg, %dreg;       \

	.align 4
	.globl	C_LABEL(csum_copy_ldo4fixup)
C_LABEL(csum_copy_ldo4fixup):
	CCOPY_FIXUP_LD(o4, o0, g6, g1)

	.align 4
	.globl	C_LABEL(csum_copy_lduho4fixup)
C_LABEL(csum_copy_lduho4fixup):
	CCOPY_FIXUP_LDUH(o4, o0, g6, g1)

	/* This macro must expand to 8-instructions==32bytes */
#define CCOPY_FIXUP_ST(sreg, breg, offset, retpc, tmp) \
	stb	%sreg, [%breg + offset + 0x3];  \
	srl	%sreg, 8, %tmp;                 \
	stb	%tmp, [%breg + offset + 0x2];   \
	srl	%sreg, 16, %tmp;                \
	stb	%tmp, [%breg + offset + 0x1];   \
	srl	%sreg, 24, %tmp;                \
	jmp	%retpc + 0x4;                   \
	 stb	%tmp, [%breg + offset + 0x0];   \

	.align 4
	.globl	C_LABEL(csum_copy_sto4fixups)
	.globl	C_LABEL(csum_copy_sto5fixups)
	.globl	C_LABEL(csum_copy_stg2fixups)
	.globl	C_LABEL(csum_copy_stg3fixups)
	.globl	C_LABEL(csum_copy_stg4fixups)
	.globl	C_LABEL(csum_copy_stg5fixups)
C_LABEL(csum_copy_sto4fixups):
	CCOPY_FIXUP_ST(o4, o1, 0x00, g6, g1) /* 0x00 */
	CCOPY_FIXUP_ST(o4, o1, 0x18, g6, g1) /* 0x20 */
	CCOPY_FIXUP_ST(o4, o1, 0x20, g6, g1) /* 0x40 */
	CCOPY_FIXUP_ST(o4, o1, 0x38, g6, g1) /* 0x60 */
	CCOPY_FIXUP_ST(o4, o1, 0x40, g6, g1) /* 0x80 */
	CCOPY_FIXUP_ST(o4, o1, 0x58, g6, g1) /* 0xa0 */
	CCOPY_FIXUP_ST(o4, o1, 0x60, g6, g1) /* 0xc0 */
	CCOPY_FIXUP_ST(o4, o1, 0x78, g6, g1) /* 0xe0 */
C_LABEL(csum_copy_sto5fixups):
	CCOPY_FIXUP_ST(o5, o1, 0x04, g6, g1) /* 0x00 */
	CCOPY_FIXUP_ST(o5, o1, 0x1c, g6, g1) /* 0x20 */
	CCOPY_FIXUP_ST(o5, o1, 0x24, g6, g1) /* 0x40 */
	CCOPY_FIXUP_ST(o5, o1, 0x3c, g6, g1) /* 0x60 */
	CCOPY_FIXUP_ST(o5, o1, 0x44, g6, g1) /* 0x80 */
	CCOPY_FIXUP_ST(o5, o1, 0x5c, g6, g1) /* 0xa0 */
	CCOPY_FIXUP_ST(o5, o1, 0x64, g6, g1) /* 0xc0 */
	CCOPY_FIXUP_ST(o5, o1, 0x7c, g6, g1) /* 0xe0 */
C_LABEL(csum_copy_stg2fixups):
	CCOPY_FIXUP_ST(g2, o1, 0x10, g6, g1) /* 0x00 */
	CCOPY_FIXUP_ST(g2, o1, 0x30, g6, g1) /* 0x20 */
	CCOPY_FIXUP_ST(g2, o1, 0x50, g6, g1) /* 0x40 */
	CCOPY_FIXUP_ST(g2, o1, 0x70, g6, g1) /* 0x60 */
C_LABEL(csum_copy_stg3fixups):
	CCOPY_FIXUP_ST(g3, o1, 0x14, g6, g1) /* 0x00 */
	CCOPY_FIXUP_ST(g3, o1, 0x34, g6, g1) /* 0x20 */
	CCOPY_FIXUP_ST(g3, o1, 0x54, g6, g1) /* 0x40 */
	CCOPY_FIXUP_ST(g3, o1, 0x74, g6, g1) /* 0x60 */
C_LABEL(csum_copy_stg4fixups):
	CCOPY_FIXUP_ST(g4, o1, 0x08, g6, g1) /* 0x00 */
	CCOPY_FIXUP_ST(g4, o1, 0x28, g6, g1) /* 0x20 */
	CCOPY_FIXUP_ST(g4, o1, 0x48, g6, g1) /* 0x40 */
	CCOPY_FIXUP_ST(g4, o1, 0x68, g6, g1) /* 0x60 */
C_LABEL(csum_copy_stg5fixups):
	CCOPY_FIXUP_ST(g5, o1, 0x0c, g6, g1) /* 0x00 */
	CCOPY_FIXUP_ST(g5, o1, 0x2c, g6, g1) /* 0x20 */
	CCOPY_FIXUP_ST(g5, o1, 0x4c, g6, g1) /* 0x40 */
	CCOPY_FIXUP_ST(g5, o1, 0x6c, g6, g1) /* 0x60 */

#define CCOPY_FIXUP_STH(sreg, breg, retpc, tmp) \
	stb	%sreg, [%breg + 0x1];  \
	srl	%sreg, 8, %tmp;        \
	jmp	%retpc + 0x4;          \
	 stb	%tmp, [%breg + 0x0];   \

	.align 4
	.globl	C_LABEL(csum_copy_stho4fixup)
C_LABEL(csum_copy_stho4fixup):
	CCOPY_FIXUP_STH(o4, o1, g6, g1)

/* yeeee.... */
