/* memcpy.S: Sparc optimized memcpy code.
 *
 *  Copyright(C) 1995 Linus Torvalds
 *  Copyright(C) 1996 David S. Miller
 *  Copyright(C) 1996 Eddie C. Dost
 *
 * derived from:
 *	e-mail between David and Eddie.
 */

#include <asm/cprefix.h>

#define MOVE_BIGCHUNK(src, dst, offset, t0, t1, t2, t3, t4, t5) \
	ldd	[%src + offset + 0x18], %t0; \
	st	%t0, [%dst + offset + 0x18]; \
	ldd	[%src + offset + 0x10], %t2; \
	st	%t1, [%dst + offset + 0x1c]; \
	ldd	[%src + offset + 0x08], %t4; \
	st	%t2, [%dst + offset + 0x10]; \
	ldd	[%src + offset + 0x00], %t0; \
	st	%t3, [%dst + offset + 0x14]; \
	st	%t0, [%dst + offset + 0x00]; \
	st	%t1, [%dst + offset + 0x04]; \
	st	%t4, [%dst + offset + 0x08]; \
	st	%t5, [%dst + offset + 0x0c]; \

	/* Alignment cases are:
	 * 1) (src&0x3)=0x0 (dst&0x3)=0x0 	can optimize
	 * 2) (src&0x3)=0x1 (dst&0x3)=0x1	can optimize
	 * 3) (src&0x3)=0x2 (dst&0x3)=0x2	can optimize
	 * 4) (src&0x3)=0x3 (dst&0x3)=0x3	can optimize
	 * 5) anything else			cannot optimize
	 */

	.text
	.align	4

	.globl	C_LABEL(bcopy), C_LABEL(amemmove), C_LABEL(memmove)
	.globl	C_LABEL(memcpy)
C_LABEL(bcopy):
	mov	%o0, %o3
	mov	%o1, %o0
	mov	%o3, %o1
C_LABEL(amemmove):
C_LABEL(memmove):
C_LABEL(memcpy):	/* %o0=dst %o1=src %o2=len */
	mov	%o0, %g6

	xor	%o0, %o1, %o4
	andcc	%o4, 3, %g0
	be,a	can_align
	 cmp	%o2, 8		! 3 might be eaten by align, leaves < 4.

	b	cannot_optimize
	 cmp	%o2, 0

can_align:
	blt	cannot_optimize
	 cmp	%o2, 0

	be	out
	 andcc	%o1, 1, %g0

hword_align:
	be	word_align
	 andcc	%o1, 2, %g0

	ldub	[%o1], %o4
	sub	%o2, 1, %o2
	stb	%o4, [%o0]
	add	%o1, 1, %o1
	add	%o0, 1, %o0
	andcc	%o1, 2, %g0

word_align:
	be	dword_align
	 cmp	%o2, 56		! 24 might be eaten by align, leaves < 32.

	lduh	[%o1], %o4
	sub	%o2, 2, %o2
	sth	%o4, [%o0]
	add	%o0, 2, %o0
	add	%o1, 2, %o1
	cmp	%o2, 56

dword_align:
	blt,a	do_end_words
	 andcc	%o2, 0x3f, %g7

	andcc	%o1, 4, %g0

	be	qword_align
	 andcc	%o1, 8, %g0

	ld	[%o1], %o4
	sub	%o2, 4, %o2
	st	%o4, [%o0]
	add	%o1, 4, %o1
	add	%o0, 4, %o0
	andcc	%o1, 8, %g0

qword_align:
	be	oword_align
	 andcc	%o1, 16, %g0

	ldd	[%o1], %o4
	sub	%o2, 8, %o2
	st	%o4, [%o0]
	add	%o1, 8, %o1
	st	%o5, [%o0 + 0x04]
	add	%o0, 8, %o0
	andcc	%o1, 16, %g0

oword_align:
	be	begin_movement
	 srl	%o2, 7, %g7

	ldd	[%o1 + 8], %g2;
	ldd	[%o1], %o4;
	st	%g2, [%o0 + 8]
	sub	%o2, 16, %o2
	st	%g3, [%o0 + 12]
	add	%o1, 16, %o1
	st	%o4, [%o0]
	add	%o0, 16, %o0
	st	%o5, [%o0 - 12]
	srl	%o2, 7, %g7

begin_movement:
	orcc	%g0, %g7, %g0
	be	0f
	 andcc	%o2, 0x40, %g0

move_128bytes:
	MOVE_BIGCHUNK(o1, o0, 0x00, o4, o5, g2, g3, g4, g5)
	MOVE_BIGCHUNK(o1, o0, 0x20, o4, o5, g2, g3, g4, g5)
	MOVE_BIGCHUNK(o1, o0, 0x40, o4, o5, g2, g3, g4, g5)
	MOVE_BIGCHUNK(o1, o0, 0x60, o4, o5, g2, g3, g4, g5)
	add	%o1, 128, %o1
	subcc	%g7, 1, %g7
	bne	move_128bytes
	 add	%o0, 128, %o0

	andcc	%o2, 0x40, %g0
0:
	be	1f
	 andcc	%o2, 0x20, %g0

move_64bytes:
	MOVE_BIGCHUNK(o1, o0, 0x00, o4, o5, g2, g3, g4, g5)
	MOVE_BIGCHUNK(o1, o0, 0x20, o4, o5, g2, g3, g4, g5)
	add	%o1, 64, %o1
	add	%o0, 64, %o0
	andcc	%o2, 0x20, %g0

1:
	be	do_end_words
	 andcc	%o2, 0x1c, %g7

move_32bytes:
	MOVE_BIGCHUNK(o1, o0, 0x00, o4, o5, g2, g3, g4, g5)
	add	%o1, 32, %o1
	add	%o0, 32, %o0

	andcc	%o2, 0x1c, %g7

do_end_words:
	be	maybe_end_cruft
	 srl	%g7, 2, %g7

end_words:
	ld	[%o1], %o4
	subcc	%g7, 1, %g7
	st	%o4, [%o0]
	add	%o1, 4, %o1
	bne	end_words
	 add	%o0, 4, %o0

maybe_end_cruft:
	andcc	%o2, 3, %o2

cannot_optimize:
	be	out
	 nop

end_bytes:
	ldub	[%o1], %o4
	subcc	%o2, 1, %o2
	stb	%o4, [%o0]
	add	%o1, 1, %o1
	bne	end_bytes
	 add	%o0, 1, %o0

out:
	retl
 	 mov	%g6, %o0


