/* $Id: cgfourteen.c,v 1.5 1996/08/04 11:36:41 ecd Exp $
 * cgfourteen.c: Sun SparcStation console support.
 *
 * Copyright (C) 1995 Miguel de Icaza (miguel@nuclecu.unam.mx)
 */

#include <linux/kd.h>
#include <linux/tty.h>
#include <linux/malloc.h>

#include <asm/sbus.h>
#include <asm/io.h>
#include <asm/fbio.h>
#include <asm/pgtable.h>

#include "../../char/vt_kern.h"
#include "../../char/selection.h"
#include "../../char/console_struct.h"
#include "fb.h"

static int
cg14_emul_mmap (struct inode *inode, struct file *file,
		struct vm_area_struct *vma, long base, fbinfo_t *fb)
{
	uint size, page, r, map_size;
	uint map_offset = 0;
	
	size = vma->vm_end - vma->vm_start;
        if (vma->vm_offset & ~PAGE_MASK)
                return -ENXIO;

	/* To stop the swapper from even considering these pages */
	vma->vm_flags |= FB_MMAP_VM_FLAGS; 
	
	/* Each page, see which map applies */
	for (page = 0; page < size; ){
		switch (vma->vm_offset+page){
		case CG3_MMAP_OFFSET-0x7000:
			printk ("Mapping fb->info.regs!\n");
			map_size = 0x7000;
			map_offset = get_phys ((uint) fb->info.cg14.regs);
			break;
			
		case CG3_MMAP_OFFSET:
			map_size = size-page;
			map_offset = get_phys ((uint) fb->base);
#if 0
			if (map_size > fb->type.fb_size)
				map_size = fb->type.fb_size;
#endif
			break;
		default:
			map_size = 0;
			break;
		}
		if (!map_size){
			page += PAGE_SIZE;
			continue;
		}
		r = io_remap_page_range (vma->vm_start+page,
					 map_offset,
					 map_size, vma->vm_page_prot,
					 fb->space);
		if (r) return -EAGAIN;
		page += map_size;
	}
        vma->vm_inode = inode;
        inode->i_count++;
        return 0;
}

static void
cg14_cmap (fbinfo_t *fb, int index, int count)
{
	struct cg14_clut *clut = fb->info.cg14.clut;
	int i;
	
	for (i = index; count--; i++){
		clut->c_clut [i] = 
			(color_map.map [i][2] << 16) |
			(color_map.map [i][1] << 8) |
			(color_map.map [i][0]);
	}
}

#ifdef hires_cg14
/* These ones are for putting the video card on 16/32 bpp */
static int
cg14_ioctl (struct inode *inode, struct file *file, unsigned cmd, unsigned long arg, fbinfo_t *fb)
{
	switch (cmd){
	case MDI_RESET: {
		volatile unsigned char *control;
		*control = (*control & ~0x30);
	}
	break;

	case MDI_GET_CFGINFO: {
		int i;
		struct mdi_cfginfo *mdii;
		
		i = verify_area (VERIFY_WRITE (void *) arg, sizeof (struct mdi_cfginfo));
		if (i) return i;
		mdii->mdi_ncluts = 2;
		mdii->mdi_type   = FBTYPE_MDICOLOR;
		mdii->mdi_height = fb->type.fb_height;
		mdii->mdi_width  = fb->type.fb_width;
		mdii->mdi_mode   = fb->info.cg14.video_mode;
		mdii->mdi_pixfreq= 72; /* FIXME */
		/* Should determine available RAM in the thing */
		mdii->mdi_size   = (fb->type.fb_width * fb->type.fb_height) * (fb->info.cg14.video_mode);
	}
	break;
	
	case FBIORES: {
		int i;
		volatile unsigned char *control;
		
		i = verify_area (VERIFY_READ, (void *) arg, sizeof (int));
		if (i) return i;
		control = fb->info.cg14.regs;
		switch (i){
		case MDI_32BPP:
			*control = (*control & ~0x30) | 0x30;
			video_mode = 4;
			break;
		case MDI_16BPP:
			*control = (*control & ~0x30) | 0x20;
			video_mode = 2;
			break;
		case MDI_8BPP:
			*control = (*control & ~0x30);
			video_mode = 1;
			break;

		default:
			return -ENOSYS;
		}
	}
	break;
	
	} /* switch */
}

static int
cg14_mmap (struct inode *inode, struct file *file, struct vm_area_struct *vma,
	   long base, fbinfo_t *fb)
{
	uint size, page, r, map_size;
	uint map_offset = 0;
	
	size = vma->vm_end - vma->vm_start;
        if (vma->vm_offset & ~PAGE_MASK)
                return -ENXIO;

	/* To stop the swapper from even considering these pages */
	vma->vm_flags |= FB_MMAP_VM_FLAGS;
	
	/* Each page, see which map applies */
	for (page = 0; page < size; ){
	}
}
#endif

void
cg14_setup (fbinfo_t *fb, int slot, uint cg14, int cg14_io, int cg14regs)
{
	struct cg14_info *cg14info;
	
	printk ("cgfourteen%d at 0x%8.8x\n", slot, cg14);
	fb->type.fb_cmsize = 256;
	fb->mmap =  cg14_emul_mmap;
	fb->loadcmap = cg14_cmap;
#ifdef hires_cg14
	fb->ioctl = cg14_ioctl;
#else
	fb->ioctl = 0;
#endif
	fb->reset = 0;
	fb->blank = 0;
	fb->unblank = 0;
	fb->info.cg14.video_mode = 0;
	cg14info = (struct cg14_info *) &fb->info.cg14;
	/* Until I know how to map the beast manually, we use PROM mapping */
	if (cg14regs){
		cg14info->regs = (void *) cg14regs;
		cg14info->clut = (void *) (cg14regs + CG14_CLUT1);
	}
	/* Lie about it */
	fb->type.fb_type = FBTYPE_SUN3COLOR;
}
