/* $Id: fb.h,v 1.13 1996/08/29 07:38:02 ecd Exp $
 * fb.h: contains the definitions of the structures that various sun
 *       frame buffer can use to do console driver stuff.
 *
 * (C) 1996 Dave Redman     (djhr@tadpole.co.uk)
 * (C) 1996 Miguel de Icaza (miguel@nuclecu.unam.mx)
 * (C) 1996 David Miller    (davem@rutgers.edu)
 * (C) 1996 Peter Zaitcev   (zaitcev@lab.ipmce.su)
 * (C) 1996 Eddie C. Dost   (ecd@skynet.be)
 */

#ifndef __SPARC_FB_H_
#define __SPARC_FB_H_

#include <linux/config.h>

#define FRAME_BUFFERS    1
#define CHAR_WIDTH	 8
#define CHAR_HEIGHT	 16

/* Change this if we run into problems if the kernel want's to free or
 * use our frame buffer pages, never seen it though.
 */
#define FB_MMAP_VM_FLAGS (VM_SHM| VM_LOCKED)

#undef color

/* cg6 cursor status, kernel tracked copy */
struct cg6_cursor {
        short   enable;	        /* cursor is enabled */
        struct  fbcurpos cpos;	/* position */
        struct  fbcurpos chot;	/* hot-spot */
        struct  fbcurpos size;	/* size of mask & image fields */
        int     bits[2][32];	/* space for mask & image bits */
	char    color [6];	/* cursor colors */
};

struct cg6_info {
	struct bt_regs *bt;	/* color control */
	struct cg6_fbc *fbc;
	unsigned int   *fhc;
	struct cg6_tec *tec;
	struct cg6_thc *thc;
	struct cg6_cursor cursor; /* cursor control */
	void *dhc;
};

struct tcx_info {
	struct bt_regs *bt;	/* color control */
	struct tcx_tec *tec;
	struct tcx_thc *thc;
	struct cg6_cursor cursor; /* cursor control */
	void *tcx_cplane;
	int tcx_sizes[13];
	long tcx_offsets[13];
	int lowdepth;
	int hw_cursor;
};

struct bwtwo_info {
        struct bwtwo_regs *regs;
};

struct cg3_info {
	struct bt_regs *bt;	/* brooktree (color) registers */
};

struct cg14_info {
    struct cg14_regs   *regs;
    struct cg14_cursor *cursor_regs;
    struct cg14_dac    *dac;
    struct cg14_xlut   *xlut;
    struct cg14_clut   *clut;
    int    video_mode;
};

typedef union
{
	unsigned int	bt[8];
	unsigned char	ibm[8];
} dacptr;
    
struct weitek_info
{
	int		p9000;		/* p9000? or p9100 */
	dacptr		*dac;		/* dac structures */
	unsigned int	fbsize;		/* size of frame buffer */
};

/* Array holding the information for the frame buffers */
typedef struct fbinfo {
	union {
		struct bwtwo_info bwtwo;
		struct cg3_info   cg3;
		struct cg6_info   cg6;
		struct cg14_info  cg14;
		struct tcx_info   tcx;
	} info;		        /* per frame information */
	int    space;           /* I/O space this card resides in */
	int    blanked;		/* true if video blanked */
	int    open;		/* is this fb open? */
	int    mmaped;		/* has this fb been mmapped? */
	int    vtconsole;	/* virtual console where it is opened */
	long   base;		/* frame buffer base    */
	struct fbtype type;	/* frame buffer type    */
	int    real_type;	/* real frame buffer FBTYPE* */
	int    emulations[4];   /* possible emulations (-1 N/A) */
	int    (*mmap)(struct inode *, struct file *, struct vm_area_struct *,
		       long fb_base, struct fbinfo *);
	void   (*loadcmap)(struct fbinfo *fb, int index, int count);
	void   (*blank)(struct fbinfo *fb);
	void   (*unblank)(struct fbinfo *fb);
	int    (*ioctl)(struct inode *, struct file *, uint, unsigned long,
			struct fbinfo *);
	void   (*reset)(struct fbinfo *fb);
	void   (*switch_from_graph)(void);
} fbinfo_t;

extern fbinfo_t fbinfo [];

struct {
	char *name;		/* prom name */
	int  width, height;	/* prefered w,h match */
	void (*fbtype)(fbinfo_t *); /* generic device type */
	/* device specific init routine  */
	unsigned long (*fbinit)(fbinfo_t *fbinfo, unsigned int addr);
} fb_entry;

extern int fb_init(void);

extern void (*fb_restore_palette)(fbinfo_t *fbinfo);
extern void (*fb_hide_cursor)(int cursor_pos);
extern void (*fb_set_cursor)(int oldpos,  int idx);
extern void (*fb_clear_screen)( void );
extern void (*fb_blitc)(unsigned char *, int, unsigned int *, unsigned int);
extern void (*fb_font_init)(unsigned char *font);
/* All framebuffers are likely to require this info */

/* Screen dimensions and color depth. */
extern int con_depth, con_width;
extern int con_height, con_linebytes;
extern int ints_per_line;

/* used in the mmap routines */
extern unsigned int get_phys (unsigned int addr);
extern void render_screen(void);

union fb_cmap
{
	unsigned char   map[256][3];    /* reasonable way to access */
	unsigned int    raw[256*3/4];   /* hardware wants it like this */
};
    
extern union fb_cmap color_map;

extern void (*hw_hide_cursor_func)(void);
extern void (*hw_set_cursor_func)(int, int);

#define FB_DEV(x) (MINOR(x) / 32)

#ifdef CONFIG_SUN_FB_GX_ACCEL
extern int gx_blitc(unsigned short charattr, int xoff, int yoff);
extern void gx_setw(int xoff, int yoff, unsigned short c, int count);
extern void gx_cpyw(int xoff, int yoff, unsigned short *p, int count);
#endif
#endif __SPARC_FB_H_
