/* $Id: sunfb.c,v 1.10 1996/08/29 07:38:08 ecd Exp $
 * sunfb.c: Sun generic frame buffer support.
 *
 * Copyright (C) 1995, 1996 Miguel de Icaza (miguel@nuclecu.unam.mx)
 * Copyright (C) 1995 David S. Miller (davem@caip.rutgers.edu)
 * 
 * Added getcmap ioctl, may, 96
 */

#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/kd.h>
#include <linux/malloc.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/types.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/bitops.h>
#include <asm/oplib.h>
#include <asm/sbus.h>
#include <asm/fbio.h>
#include <asm/io.h>
#include <asm/pgtsun4c.h>	/* for the sun4c_nocache */

#include "../../char/kbd_kern.h"
#include "../../char/vt_kern.h"
#include "../../char/consolemap.h"
#include "../../char/selection.h"
#include "../../char/console_struct.h"

#include "fb.h"

extern void set_palette (void);
extern void sun_clear_screen(void);

static int
fb_open (struct inode * inode, struct file * file)
{
	int minor = FB_DEV (inode->i_rdev);

	if (minor >= FRAME_BUFFERS)
		return -EBADF;
	if (fbinfo [minor].open)
		return -EBUSY;
	fbinfo [minor].open = 1;
	fbinfo [minor].mmaped = 0;
	return 0;
}

static int
fb_ioctl (struct inode *inode, struct file *file, uint cmd, unsigned long arg)
{
	int minor = FB_DEV (inode->i_rdev);
	fbinfo_t *fb;
	struct fbcmap *cmap;
	int i;
	
	if (minor >= FRAME_BUFFERS)
		return -EBADF;
	fb = &fbinfo [minor];
	
	switch (cmd){
	case FBIOGTYPE:		/* return frame buffer type */
		i = verify_area (VERIFY_WRITE, (void *) arg, sizeof (struct fbtype));
		if (i) return i;
		*(struct fbtype *)arg = (fb->type);
		break;
	case FBIOGATTR:{
		struct fbgattr *fba = (struct fbgattr *) arg;
		
		i = verify_area (VERIFY_WRITE, (void *) arg, sizeof (struct fbgattr));
		if (i) return i;
		fba->real_type = fb->real_type;
		fba->owner = 0;
		fba->fbtype = fb->type;
		fba->sattr.flags = 0;
		fba->sattr.emu_type = fb->type.fb_type;
		fba->sattr.dev_specific [0] = -1;
		fba->emu_types [0] = fb->type.fb_type;
		for (i = 1; i < 4; i++)
			fba->emu_types [i] = fb->emulations [i];
		break;
	}
	case FBIOSATTR:
		i = verify_area (VERIFY_READ, (void *) arg, sizeof (struct fbsattr));
		if (i) return i;
		return -EINVAL;
	case FBIOSVIDEO:
		i = verify_area(VERIFY_READ, (void *)arg, sizeof(int));
		if (i) return i;
		
		if (*(int *)arg){
			if (!fb->blanked || !fb->unblank)
				break;
			(*fb->unblank)(fb);
			fb->blanked = 0;
		} else {
			if (fb->blanked || !fb->blank)
				break;
			(*fb->blank)(fb);
			fb->blanked = 1;
		}
		break;
	case FBIOGVIDEO:
		i = verify_area (VERIFY_WRITE, (void *) arg, sizeof (int));
		if (i) return i;
		*(int *) arg = fb->blanked;
		break;
	case FBIOGETCMAP: {
		char *rp, *gp, *bp;
		int end, count;;

		if (!fb->loadcmap)
			return -EINVAL;
		i = verify_area (VERIFY_READ, (void *) arg, sizeof (struct fbcmap));
		if (i) return i;
		cmap = (struct fbcmap *) arg;
		count = cmap->count;
		if ((cmap->index < 0) || (cmap->index > 255))
			return -EINVAL;
		if (cmap->index + count > 256)
			count = 256 - cmap->index;
		i = verify_area (VERIFY_WRITE, rp = cmap->red, cmap->count);
		if (i) return i;
		i = verify_area (VERIFY_WRITE, gp = cmap->green, cmap->count);
		if (i) return i;
		i = verify_area (VERIFY_WRITE, bp = cmap->blue, cmap->count);
		if (i) return i;
		end = cmap->index + count;
		for (i = cmap->index; i < end; i++){
			*rp++ = color_map.map [i][0];
			*gp++ = color_map.map [i][1];
			*bp++ = color_map.map [i][2];
		}
		(*fb->loadcmap)(fb, cmap->index, count);
                break;			

	}
	case FBIOPUTCMAP: {	/* load color map entries */
		char *rp, *gp, *bp;
		int end, count;;
		
		if (!fb->loadcmap)
			return -EINVAL;
		i = verify_area (VERIFY_READ, (void *) arg, sizeof (struct fbcmap));
		if (i) return i;
		cmap = (struct fbcmap *) arg;
		count = cmap->count;
		if ((cmap->index < 0) || (cmap->index > 255))
			return -EINVAL;
		if (cmap->index + count > 256)
			count = 256 - cmap->index;
		i = verify_area (VERIFY_READ, rp = cmap->red, cmap->count);
		if (i) return i;
		i = verify_area (VERIFY_READ, gp = cmap->green, cmap->count);
		if (i) return i;
		i = verify_area (VERIFY_READ, bp = cmap->blue, cmap->count);
		if (i) return i;

		end = cmap->index + count;
		for (i = cmap->index; i < end; i++){
			color_map.map [i][0] = *rp++;
			color_map.map [i][1] = *gp++;
			color_map.map [i][2] = *bp++;
		}
		(*fb->loadcmap)(fb, cmap->index, count);
                break;			
	}

	default:
		if (fb->ioctl){
			i = fb->ioctl (inode, file, cmd, arg, fb);
			if (i == -ENOSYS) {
				printk ("[[FBIO: %8.8x]]\n", cmd);
				return -EINVAL;
			}
			return i;
		}
		printk ("[[FBIO: %8.8x]]\n", cmd);
		return -EINVAL;
	}
	return 0;
}

static void
fb_close (struct inode * inode, struct file *filp)
{
	int minor = FB_DEV (inode->i_rdev);
	fbinfo_t *fb;
	struct fbcursor cursor;
	
	if (minor >= FRAME_BUFFERS)
		return;

	fb = &fbinfo[minor];
	if (fb->open)
		fb->open = 0;
	vt_cons [fb->vtconsole]->vc_mode = KD_TEXT;

	/* Leaving graphics mode, turn off the cursor */
	if (fb->mmaped)
		sun_clear_screen ();
	cursor.set    = FB_CUR_SETCUR;
	cursor.enable = 0;
	
	/* Reset the driver */
	if (fb->reset)
		fb->reset(fb);

	fb_ioctl (inode, filp, FBIOSCURPOS, (unsigned long) &cursor);
	set_palette ();
	render_screen ();
	return;
}

static int
fb_mmap (struct inode *inode, struct file *file, struct vm_area_struct *vma)
{
	int minor = FB_DEV (inode->i_rdev);
	fbinfo_t *fb;

	if (minor >= FRAME_BUFFERS)
		return -ENXIO;

	fb = &fbinfo [minor];

	if (fb->mmap){
		int v;
		
		v = (*fb->mmap)(inode, file, vma, fb->base, fb);
		if (v) return v;
		if (!fb->mmaped) {
			fb->mmaped = 1;
			fb->vtconsole = fg_console;
			vt_cons [fg_console]->vc_mode = KD_GRAPHICS;
		}
		return 0;
	} else
		return -ENXIO;
}

static struct file_operations graphdev_fops =
{
	NULL,			/* lseek */
	NULL,			/* read */
	NULL,			/* write */
	NULL,			/* readdir */
	NULL,			/* select */
	fb_ioctl,
	fb_mmap,
	fb_open,		/* open */
	fb_close,		/* close */
};

int
fb_init (void)
{
	/* Register the frame buffer device */
	if (register_chrdev (GRAPHDEV_MAJOR, "graphics", &graphdev_fops)){
		printk ("Could not register graphics device\n");
		return -EIO;
	}
	return 0; /* success */
}    
